#define F_CPU 16000000
#define F_CPU_div F_CPU/1000000

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#include "usart.h"

#define IN1         PC0
#define IN2         PC1
#define IN3         PC2
#define IN4         PC3
#define SPEED       200
// Trigger-Echo pairs for the sensors
#define FRONTT      PA3
#define LEFTT       PA4
#define RIGHTT      PA5
#define FRONTE       PA0
#define LEFTE        PA1
#define RIGHTE       PA2

static volatile int overflows = 0;

// Number of overflows for the sensors
ISR(TIMER0_OVF_vect)
{
    overflows++;
}

void PORTS_init(void)
{
    DDRD = 0;
    DDRD |= (1 << PD4) | (1 << PD5) | (1 << PD7);

    DDRA |= (1 << FRONTT) | (1 << LEFTT) | (1 << RIGHTT);
    DDRA &= ~(1 << FRONTT) | ~(1 << LEFTT) | ~(1 << RIGHTT);
}

void init_engine(void)
{
    PORTC = 0;
    DDRC = 0;
    DDRC |= (1 << IN1) | (1 << IN3);
    PORTC |= (1 << IN1) | (1 << IN3);
}

void TIMER0_init()
{   
    // Activating overflow ISR for Timer0
    TIMSK0 |= (1 << TOIE0);
}

void TIMER1_init()
{
    //FastPWM with 1024 prescaler
    TCCR1A |= (1 << COM1A1) | (1 << COM1B1) | (3 << WGM10);
    TCCR1B |= (1 << WGM12) | (1 << CS12) | (1 << CS10);

    OCR1A = SPEED;
    OCR1B = SPEED;
}


void move_front(void)
{
    OCR1A = SPEED + 31;
    OCR1B = SPEED;
}

void stop_engine(int seconds)
{
    OCR1A = 0;
    OCR1B = 0;
    _delay_ms(seconds * 1000);
    OCR1A = SPEED;
    OCR2A = SPEED;
}

//turns 90 degrees
void turn_right()
{
    //set direction
    PORTC &= ~_BV(IN3);
    PORTC |= _BV(IN4);

    OCR1A = SPEED;
    OCR1B = SPEED;
    _delay_ms(2500);
    
    //set direction back to normal
    PORTC &= ~_BV(IN4);
    PORTC |= _BV(IN3);

    stop_engine(2);
    
}

//turns 90 degrees
void turn_left()
{
    //set direction
    PORTC &= ~_BV(IN1);
    PORTC |= _BV(IN2);
    
    OCR1A = SPEED;
    OCR1B = SPEED;
    _delay_ms(2500);
    
    //set direction back to normal
    PORTC &= ~_BV(IN2);
    PORTC |= _BV(IN1);
    
    stop_engine(2);
}

int mesasure_distance(char sensor) {
    int trig = FRONTT, echo = FRONTT;
    int pulse = 0;

    overflows = 0;
    switch(sensor) {
        case 'F':
            trig = FRONTT;
            echo = FRONTE;
            break;
        case 'L':
            trig = LEFTT;
            echo = LEFTE;
            break;
        case 'R':
            trig = RIGHTT;
            echo = RIGHTT;
            break;
    }

    // Setting the trigger high then immediately after low
    PORTA |= (1 << trig);
    _delay_ms(10);
    PORTA &= ~(1 << trig);
    
    // Waiting while the echo is 0
    while(!(PINA & (1 << echo)));
    
    // Starting the timer
    TCCR0B |= (1 << CS10);
    TCNT0 = 0;
    
    // Waiting for the echo to go back at 0
    while(PINA & (1 << echo));
    
    pulse = TCNT0;

    // Computing the distance
    return (pulse + 255 * overflows) / (F_CPU_div * 58);
}

int check_between(int val, int min, int max) {
    if(val > min && val < max) {
        return 1;
    }

    return 0;
}

void movement_logic()
{
    int dist_front, dist_left, dist_right;
   

    while(1) {
        dist_front = mesasure_distance('F');
        dist_left = mesasure_distance('L');
        dist_right = mesasure_distance('R');
        
        // The flow
        // If theres a wall in the front check left
        // If theres a wall in the left check right
        // If theres a wall in the right turn 180 and hope for the best
        // If theres no wall right, turn right
        // If theres no wall left, turn left
        // If theres no wall in the front, keep going boy
        move_front();
        if(check_between(dist_front, 0, 10)) {
            if(check_between(dist_left, 0, 10)) {
                if(check_between(dist_right, 0, 10)) {
                    turn_left();
                    turn_left();
                } else {
                    turn_right();
                }
            } else {
                turn_left();
            }
        } else {
            move_front();
        }

    }
}



int main(void) {
	sei();
    
    PORTS_init();
    DDRD |= _BV(PD7);    
    init_engine();

	USART0_init();

    TIMER0_init();
	TIMER1_init();

    movement_logic();

	return 0;
}
